"""
RabbitMQ
============

https://www.rabbitmq.com/

"""
import sys

from fabric.api import settings

from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

DJANGO = 'django'
LOCAL = 'local'

class RabbitMQSatchel(ServiceSatchel):

    name = 'rabbitmq'

    ## Service options.

    ignore_errors = True

    @property
    def packager_system_packages(self):
        return {
            FEDORA: ['rabbitmq-server'],
            UBUNTU: ['rabbitmq-server'],
            (UBUNTU, '12.04'): ['rabbitmq-server'],
            (UBUNTU, '14.04'): ['rabbitmq-server'],
            (UBUNTU, '16.04'): ['rabbitmq-server'],
        }

    def set_defaults(self):
        self.env.rabbitmq_host = "localhost"
        self.env.vhost = "/"
        self.env.erlang_cookie = ''
        self.env.nodename = "rabbit"
        self.env.user = "guest" # DEPRECATED
        self.env.password = "guest" # DEPRECATED
        self.env.node_ip_address = ''
        self.env.port = 5672
        self.env.erl_args = ""
        self.env.cluster = "no"
        self.env.cluster_config = "/etc/rabbitmq/rabbitmq_cluster.config"
        self.env.logdir = "/var/log/rabbitmq"
        self.env.mnesiadir = "/var/lib/rabbitmq/mnesia"
        self.env.start_args = ""
        self.env.erlang_cookie_template = ''
        self.env.ignore_service_errors = 0
        self.env.management_enabled = False
        self.env.loopback_users = False
        self.env.admin_username = 'admin'
        self.env.user_lookup_method = None # LOCAL|DJANGO
        self.env.users_vhosts = [] # [(user, password, vhost)]

        self.env.auto_purge_mnesia_enabled = False
        self.env.auto_purge_mnesia_user = 'root'
        self.env.auto_purge_mnesia_command_template = 'rabbitmq/purge_mnesia.sh.template'
        self.env.auto_purge_mnesia_command_path = '/usr/local/bin/purge_mnesia.sh'
        self.env.auto_purge_mnesia_crontab_template = 'rabbitmq/etc_crond_purge_mnesia.template'
        self.env.auto_purge_mnesia_crontab_path = '/etc/cron.d/purge_mnesia'
        self.env.auto_purge_mnesia_cron_schedule = '0 0 * * SAT' # Every Saturday at midnight.
        self.env.auto_purge_mnesia_post_script = None

        # If true, enables a third-party reposistory to install the most recent version.
        self.env.bleeding = False

        self.env.service_commands = {
            START:{
                FEDORA: 'systemctl start rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server start',
            },
            STOP:{
                FEDORA: 'systemctl stop rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server stop',
            },
            DISABLE:{
                FEDORA: 'systemctl disable rabbitmq-server.service',
                UBUNTU: 'chkconfig rabbitmq-server off',
            },
            ENABLE:{
                FEDORA: 'systemctl enable rabbitmq-server.service',
                UBUNTU: 'chkconfig rabbitmq-server on',
            },
            RESTART:{
                FEDORA: 'systemctl restart rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server restart; sleep 5',
            },
            STATUS:{
                FEDORA: 'systemctl status rabbitmq-server.service',
                UBUNTU: 'service rabbitmq-server status | cat',
                (UBUNTU, '14.04'): 'service rabbitmq-server status',
                (UBUNTU, '16.04'): 'service rabbitmq-server status | cat',
            },
        }

    @task
    def configure_bleeding(self):
        """
        Enables the repository for a most current version on Debian systems.

            https://www.rabbitmq.com/install-debian.html
        """
        lm = self.last_manifest
        r = self.local_renderer
        if self.env.bleeding and not lm.bleeding:
            # Install.
            r.append(
                text='deb http://www.rabbitmq.com/debian/ testing main',
                filename='/etc/apt/sources.list.d/rabbitmq.list',
                use_sudo=True)
            r.sudo('cd /tmp; wget -O- https://www.rabbitmq.com/rabbitmq-release-signing-key.asc | sudo apt-key add -')
            r.sudo('apt-get update')

        elif not self.env.bleeding and lm.bleeding:
            # Uninstall.
            r.sudo('rm -f /etc/apt/sources.list.d/rabbitmq.list')
            r.sudo('apt-get update')

    def render_paths(self):
        r = self.local_renderer
        if self.env.erlang_cookie_template:
            r.env.erlang_cookie = r.format(self.env.erlang_cookie_template)

    @task
    def list_vhosts(self):
        """
        Displays a list of configured RabbitMQ vhosts.
        """
        self.sudo('rabbitmqctl list_vhosts')

    @task
    def list_users(self):
        """
        Displays a list of configured RabbitMQ users.
        """
        self.sudo('rabbitmqctl list_users')

    @task
    def list_queues(self):
        self.sudo('rabbitmqctl list_queues')

    @task
    def create_user(self, username, password):
        r = self.local_renderer
        r.env._user = username
        r.env._password = password
        with self.settings(warn_only=True):
            r.sudo('rabbitmqctl add_user {_user} {_password}')

    @task
    def force_stop_and_purge(self):
        """
        Forcibly kills Rabbit and purges all its queues.

        For emergency use when the server becomes unresponsive, even to service stop calls.

        If this also fails to correct the performance issues, the server may have to be completely
        reinstalled.
        """
        r = self.local_renderer
        self.stop()
        with settings(warn_only=True):
            r.sudo('killall rabbitmq-server')
        with settings(warn_only=True):
            r.sudo('killall beam.smp')
        #TODO:explicitly delete all subfolders, star-delete doesn't work
        r.sudo('rm -Rf /var/lib/rabbitmq/mnesia/*')

    @task
    def add_admin_user(self):
        r = self.local_renderer
        with self.settings(warn_only=True):
            r.sudo('rabbitmqctl add_user {admin_username} {admin_password}')
        r.sudo('rabbitmqctl set_user_tags {admin_username} administrator')
        r.sudo('rabbitmqctl set_permissions -p / {admin_username} ".*" ".*" ".*"')

    @task
    def enable_management_interface(self):
        r = self.local_renderer
        r.sudo('rabbitmq-plugins enable rabbitmq_management')
        r.sudo('service rabbitmq-server restart; sleep 3')
        self.add_admin_user()
        print('You should now be able to access the RabbitMQ web console from:')
        print('\n    http://%s:15672/' % self.genv.host_string)

    @task
    def set_loopback_users(self):
        # This allows guest to login through the management interface.
        self.sudo('touch /etc/rabbitmq/rabbitmq.config')
        #self.sudo("echo '[{rabbit, [{loopback_users, []}]}].' >> /etc/rabbitmq/rabbitmq.config")
        self.append(filename='/etc/rabbitmq/rabbitmq.config', text='[{rabbit, [{loopback_users, []}]}].', use_sudo=True)
        with self.settings(warn_only=True):
            self.sudo('service rabbitmq-server restart; sleep 3;')

    def get_user_vhosts(self, site=None):
        params = set() # [(user, password, vhost)]
        site = site or ALL
        if self.verbose:
            print('Using looking method:', self.env.user_lookup_method)
        if self.env.user_lookup_method == DJANGO:
            # Retrieve user settings from one or more assoicated Django sites.
            dj = self.get_satchel('dj')
            for _site, site_data in self.iter_sites(site=site, renderer=self.render_paths, no_secure=True):
                if self.verbose:
                    print('!'*80, file=sys.stderr)
                    print('site:', _site, file=sys.stderr)

                _settings = dj.get_settings(site=_site)
                if not _settings:
                    continue

                if hasattr(_settings, 'BROKER_USER') and hasattr(_settings, 'BROKER_VHOST'):
                    if self.verbose:
                        print('RabbitMQ:', _settings.BROKER_USER, _settings.BROKER_VHOST)
                    params.add((_settings.BROKER_USER, _settings.BROKER_PASSWORD, _settings.BROKER_VHOST))

        elif self.env.user_lookup_method == LOCAL:
            # Retrieve user settings from our local settings.
            params.update(tuple(_) for _ in self.env.users_vhosts)

        elif self.env.user_lookup_method:
            raise NotImplementedError('Unknown user lookup method: %s' % self.env.user_lookup_method)

        return params

    @task
    def install_purge_script(self):
        r = self.local_renderer

        with self.capture_bash() as capture:
            self._configure_users()

        r.env.user_script = capture.stdout.getvalue()

        # Install script to perform the actual check.
        self.install_script(
            local_path=r.env.auto_purge_mnesia_command_template,
            remote_path=r.env.auto_purge_mnesia_command_path,
            render=True,
            extra=r.collect_genv())
        r.sudo('chown root:root {auto_purge_mnesia_command_path}')

        # Install crontab to schedule running the script.
        self.install_script(
            local_path=r.env.auto_purge_mnesia_crontab_template,
            remote_path=r.env.auto_purge_mnesia_crontab_path,
            render=True,
            extra=r.collect_genv())
        r.sudo('chown root:root {auto_purge_mnesia_crontab_path}')
        r.sudo('chmod 600 {auto_purge_mnesia_crontab_path}')
        r.sudo('service cron restart')

    @task
    def purge_mnesia(self, *args, **kwargs):
        """
        Deletes all persistent data on the server.
        """
        self.install_purge_script()
        r = self.local_renderer
        s_args = (' '.join(map(str, args))) + ' ' + (' '.join('%s=%s' % (k, v) for k, v in kwargs.items()))
        s_args = s_args.strip()
        r.sudo('{auto_purge_mnesia_command_path}' + ((' ' + s_args) if s_args else ''))

    @task
    def uninstall_purge_script(self):
        r = self.local_renderer
        r.sudo('rm -f {auto_purge_mnesia_command_path}')
        r.sudo('rm -f {auto_purge_mnesia_crontab_path}')
        r.sudo('service cron restart')

    def _configure_users(self, site=None, full=0, only_data=0):
        """
        Installs and configures RabbitMQ.
        """

        site = site or ALL

        full = int(full)

        if full and not only_data:
            packager = self.get_satchel('packager')
            packager.install_required(type=SYSTEM, service=self.name)

        r = self.local_renderer

        params = self.get_user_vhosts(site=site) # [(user, password, vhost)]

        with settings(warn_only=True):
            self.add_admin_user()

        params = sorted(list(params))
        if not only_data:
            for user, password, vhost in params:
                r.env.broker_user = user
                r.env.broker_password = password
                r.env.broker_vhost = vhost
                with settings(warn_only=True):
                    r.sudo('rabbitmqctl add_user {broker_user} {broker_password}')
                    r.sudo('rabbitmqctl add_vhost {broker_vhost}')
                    r.sudo('rabbitmqctl set_permissions -p {broker_vhost} {broker_user} ".*" ".*" ".*"')
                    r.sudo('rabbitmqctl set_permissions -p {broker_vhost} {admin_username} ".*" ".*" ".*"')

        return params

    def record_manifest(self):
        """
        Returns a dictionary representing a serialized state of the service.
        """
        data = super().record_manifest()
        params = sorted(list(self.get_user_vhosts())) # [(user, password, vhost)]
        data['rabbitmq_all_site_vhosts'] = params
        data['sites'] = list(self.genv.sites or [])
        return data

    @task(precursors=['packager', 'user'])
    def configure(self, site=None, **kwargs):
        lm = self.last_manifest

        self.configure_bleeding()

        self.start()

        if self.env.auto_purge_mnesia_enabled:
            self.install_purge_script()
        elif lm.auto_purge_mnesia_enabled:
            self.uninstall_purge_script()

        if self.env.management_enabled != lm.management_enabled:
            self.enable_management_interface()

        if self.env.loopback_users != lm.loopback_users:
            self.set_loopback_users()

        return self._configure_users(**kwargs)

rabbitmq = RabbitMQSatchel()



import os
import socket
from pprint import pprint

import six

from fabric.api import (
    env,
    require,
    runs_once,
)

from burlap import common
from burlap import constants as c
from burlap.decorators import task_or_dryrun

try:
    import boto
    import boto.ec2
except ImportError:
    boto = None

EC2 = 'ec2'
KVM = 'kvm'

if 'vm_name_tag' not in env:

    env.vm_name_tag = 'Name'
    env.vm_group_tag = 'Group'
    env.vm_release_tag = 'Release'

    env.vm_type = None

    # If a name is not given, one will be auto-generated based on this pattern.
    env.vm_name_template = 'web{index}'

    # A release tag given to the instance when created to distinquish it from
    # future upgrades to the same instance name.
    env.vm_release = None

    env.vm_ec2_account_id = None
    # https://help.ubuntu.com/community/EC2StartersGuide#Official_Ubuntu_Cloud_Guest_Amazon_Machine_Images_.28AMIs.29
    env.vm_ec2_ami = None # e.g. 'ami-a29943cb'
    env.vm_ec2_instance_type = None # e.g. 'm1.small'
    env.vm_ec2_ebs = None
    env.vm_ec2_region = None # e.g. 'us-east-1'
    env.vm_ec2_zone = None # e.g. 'us-east-1b'
    env.vm_ec2_available_security_groups = {} # {(name,desc):[(protocol, port, port, ip_range)]
    env.vm_ec2_selected_security_groups = []
    env.vm_ec2_aws_access_key_id = None
    env.vm_ec2_aws_secret_access_key = None
    env.vm_ec2_volume = '/dev/sdh1'
    env.vm_ec2_keypair_name = None
    env.vm_ec2_use_elastic_ip = False
    env.vm_ec2_subnet_id = None
    env.vm_ec2_allocate_address_domain = None

    # If true, we will attempt to add or delete group rules.
    env.vm_ec2_security_group_owner = False

    # Stores dynamically allocated EIP for each host, {hostname: ip}.
    # Usually stored in a shelf file.
    env.vm_elastic_ip_mappings = None

def retrieve_ec2_hosts(extended=0, site=None):
    verbose = common.get_verbose()
    extended = int(extended)
    if verbose:
        print('site:', site)
    for host_name, data in list_instances(show=0, verbose=verbose).items():
        if verbose:
            print('host_name:', host_name)
            pprint(data, indent=4)

        # Ignore hosts that are disabled for the given site.
        if site not in (None, c.ALL) and env.available_sites_by_host and host_name in env.available_sites_by_host:
            if site not in env.available_sites_by_host[host_name]:
                if verbose:
                    print('skipping because site %s is not set for this host' % site)
                continue

        if extended:
            yield (host_name, data)
        elif data.public_dns_name:
            yield data.public_dns_name
        else:
            yield data.ip

env.hosts_retrievers[EC2] = retrieve_ec2_hosts

def translate_ec2_hostname(hostname):
    verbose = common.get_verbose()
    for name, data in list_instances(show=0, verbose=verbose).items():
        if name == hostname:
            return data.public_dns_name

env.hostname_translators[EC2] = translate_ec2_hostname

def get_ec2_connection():
    conn = boto.ec2.connect_to_region(
        #env.vm_ec2_zone,
        env.vm_ec2_region,
        aws_access_key_id=env.vm_ec2_aws_access_key_id,
        aws_secret_access_key=env.vm_ec2_aws_secret_access_key,
    )
    assert conn, 'Unable to create EC2 connection with region %s and access key %s.' % (env.vm_ec2_region, env.vm_ec2_aws_access_key_id)
    return conn

def get_all_ec2_instances(instance_ids=None):
    conn = get_ec2_connection()
    #return sum(map(lambda r: r.instances, conn.get_all_instances(instance_ids=instance_ids)), [])
    return sum([r.instances for r in conn.get_all_instances(instance_ids=instance_ids)], [])

def get_all_running_ec2_instances():
    #instances = filter(lambda i: i.state == 'running', get_all_ec2_instances())
    instances = [i for i in get_all_ec2_instances() if i.state == 'running']
    instances.reverse()
    return instances



@task_or_dryrun
#@runs_once #breaks get_or_create()
def list_instances(show=1, name=None, group=None, release=None, except_release=None):
    """
    Retrieves all virtual machines instances in the current environment.
    """
    from burlap.common import shelf, OrderedDict, get_verbose

    verbose = get_verbose()
    require('vm_type', 'vm_group')
    assert env.vm_type, 'No VM type specified.'
    env.vm_type = (env.vm_type or '').lower()
    _name = name
    _group = group
    _release = release
    if verbose:
        print('name=%s, group=%s, release=%s' % (_name, _group, _release))

    env.vm_elastic_ip_mappings = shelf.get('vm_elastic_ip_mappings')

    data = type(env)()
    if env.vm_type == EC2:
        if verbose:
            print('Checking EC2...')
        for instance in get_all_running_ec2_instances():
            name = instance.tags.get(env.vm_name_tag)
            group = instance.tags.get(env.vm_group_tag)
            release = instance.tags.get(env.vm_release_tag)
            if env.vm_group and env.vm_group != group:
                if verbose:
                    print(('Skipping instance %s because its group "%s" '
                        'does not match env.vm_group "%s".') \
                            % (instance.public_dns_name, group, env.vm_group))
                continue
            if _group and group != _group:
                if verbose:
                    print(('Skipping instance %s because its group "%s" '
                        'does not match local group "%s".') \
                            % (instance.public_dns_name, group, _group))
                continue
            if _name and name != _name:
                if verbose:
                    print(('Skipping instance %s because its name "%s" '
                        'does not match name "%s".') \
                            % (instance.public_dns_name, name, _name))
                continue
            if _release and release != _release:
                if verbose:
                    print(('Skipping instance %s because its release "%s" '
                        'does not match release "%s".') \
                            % (instance.public_dns_name, release, _release))
                continue
            if except_release and release == except_release:
                continue
            if verbose:
                print('Adding instance %s (%s).' % (name, instance.public_dns_name or instance.ip_address))
            data.setdefault(name, type(env)())
            data[name]['id'] = instance.id
            data[name]['public_dns_name'] = instance.public_dns_name or instance.ip_address
            if verbose:
                print('Public DNS: %s' % instance.public_dns_name)

            if env.vm_elastic_ip_mappings and name in env.vm_elastic_ip_mappings:
                data[name]['ip'] = env.vm_elastic_ip_mappings[name]
            else:
                data[name]['ip'] = socket.gethostbyname(instance.public_dns_name)

        if int(show):
            pprint(data, indent=4)
        return data
    if env.vm_type == KVM:
        #virsh list
        pass
    else:
        raise NotImplementedError

#@task_or_dryrun
#@runs_once
#def list(*args, **kwargs):
#    #execute(list_instances, *args, **kwargs)
#    list_instances(*args, **kwargs)


def set_ec2_security_group_id(name, id): # pylint: disable=redefined-builtin
    from burlap.common import shelf, OrderedDict
    v = shelf.get('vm_ec2_security_group_ids', OrderedDict())
    v[name] = str(id)
    shelf.set('vm_ec2_security_group_ids', v)


@task_or_dryrun
def get_ec2_security_group_id(name=None, verbose=0):
    from burlap.common import shelf, OrderedDict

    verbose = int(verbose)

    group_id = None
    conn = get_ec2_connection()
    groups = conn.get_all_security_groups()
    for group in groups:
        if verbose:
            print('group:', group.name, group.id)
        if group.name == name:
            group_id = group.id

    # Otherwise try the local cache.
    if not group_id:
        v = shelf.get('vm_ec2_security_group_ids', OrderedDict())
        group_id = v.get(name)

    if verbose:
        print(group_id)
    return group_id


@task_or_dryrun
def get_or_create_ec2_security_groups(names=None, verbose=1):
    """
    Creates a security group opening 22, 80 and 443
    """
    verbose = int(verbose)

    if verbose:
        print('Creating EC2 security groups...')

    conn = get_ec2_connection()

    if isinstance(names, six.string_types):
        names = names.split(',')
    names = names or env.vm_ec2_selected_security_groups
    if verbose:
        print('Group names:', names)

    ret = []
    for name in names:
        try:
            group_id = get_ec2_security_group_id(name)
            if verbose:
                print('group_id:', group_id)
            #group = conn.get_all_security_groups(groupnames=[name])[0]
            # Note, groups in a VPC can't be referred to by name?
            group = conn.get_all_security_groups(group_ids=[group_id])[0]
        except boto.exception.EC2ResponseError as e:
            if verbose:
                print(e)
            group = get_ec2_connection().create_security_group(
                name,
                name,
                vpc_id=env.vm_ec2_vpc_id,
            )
            print('group_id:', group.id)
            set_ec2_security_group_id(name, group.id)
        ret.append(group)

        # Find existing rules.
        actual_sets = set()
        for rule in list(group.rules):
            ip_protocol = rule.ip_protocol
            from_port = rule.from_port
            to_port = rule.to_port
            for cidr_ip in rule.grants:
                #print('Revoking:', ip_protocol, from_port, to_port, cidr_ip)
                #group.revoke(ip_protocol, from_port, to_port, cidr_ip)
                rule_groups = ((rule.groups and rule.groups.split(',')) or [None])
                for src_group in rule_groups:
                    src_group = (src_group or '').strip()
                    if src_group:
                        actual_sets.add((ip_protocol, from_port, to_port, str(cidr_ip), src_group))
                    else:
                        actual_sets.add((ip_protocol, from_port, to_port, str(cidr_ip)))

        # Find actual rules.
        expected_sets = set()
        for authorization in env.vm_ec2_available_security_groups.get(name, []):
            if verbose:
                print('authorization:', authorization)
            if len(authorization) == 4 or (len(authorization) == 5 and not (authorization[-1] or '').strip()):
                src_group = None
                ip_protocol, from_port, to_port, cidr_ip = authorization[:4]
                if cidr_ip:
                    expected_sets.add((ip_protocol, str(from_port), str(to_port), cidr_ip))
            else:
                ip_protocol, from_port, to_port, cidr_ip, src_group = authorization
                if cidr_ip:
                    expected_sets.add((ip_protocol, str(from_port), str(to_port), cidr_ip, src_group))

        # Calculate differences and update rules if we own the group.
        if env.vm_ec2_security_group_owner:
            if verbose:
                print('expected_sets:')
                print(expected_sets)
                print('actual_sets:')
                print(actual_sets)
            del_sets = actual_sets.difference(expected_sets)
            if verbose:
                print('del_sets:')
                print(del_sets)
            add_sets = expected_sets.difference(actual_sets)
            if verbose:
                print('add_sets:')
                print(add_sets)

            # Revoke deleted.
            for auth in del_sets:
                print(len(auth))
                print('revoking:', auth)
                group.revoke(*auth)

            # Create fresh rules.
            for auth in add_sets:
                print('authorizing:', auth)
                group.authorize(*auth)

    return ret

@task_or_dryrun
def get_or_create_ec2_key_pair(name=None, verbose=1):
    """
    Creates and saves an EC2 key pair to a local PEM file.
    """
    verbose = int(verbose)
    name = name or env.vm_ec2_keypair_name
    pem_path = 'roles/%s/%s.pem' % (env.ROLE, name)
    conn = get_ec2_connection()
    kp = conn.get_key_pair(name)
    if kp:
        print('Key pair %s already exists.' % name)
    else:
        # Note, we only get the private key during creation.
        # If we don't save it here, it's lost forever.
        kp = conn.create_key_pair(name)
        open(pem_path, 'wb').write(kp.material)
        os.system('chmod 600 %s' % pem_path)
        print('Key pair %s created.' % name)
    #return kp
    return pem_path

@task_or_dryrun
def list_security_groups():
    conn = get_ec2_connection()
    sgs = conn.get_all_security_groups()
    print('Id,Name,Number of Instances')
    for sg in sorted(sgs, key=lambda o: o.name):
        print('%s,%s,%s' % (sg.id, sg.name, len(sg.instances())))

@task_or_dryrun
def exists(name=None, group=None, release=None, except_release=None, verbose=1):
    """
    Determines if a virtual machine instance exists.
    """
    verbose = int(verbose)
    instances = list_instances(
        name=name,
        group=group,
        release=release,
        except_release=except_release,
        verbose=verbose,
        show=verbose)
    ret = bool(instances)
    if verbose:
        print('\ninstance %s exist' % ('DOES' if ret else 'does NOT'))
    #return ret
    return instances

@task_or_dryrun
def get_name():
    """
    Retrieves the instance name associated with the current host string.
    """
    if env.vm_type == EC2:
        for instance in get_all_running_ec2_instances():
            if env.host_string == instance.public_dns_name:
                name = instance.tags.get(env.vm_name_tag)
                return name
    else:
        raise NotImplementedError

@task_or_dryrun
def shutdown(force=False):
    #virsh shutdown <name>
    #virsh destroy <name> #to force
    raise NotImplementedError

@task_or_dryrun
def reboot():
    #virsh reboot <name>
    raise NotImplementedError

@task_or_dryrun
@runs_once
def list_ips():
    data = list_instances(show=0, verbose=0)
    for key, attrs in data.items():
        print(attrs.get('ip'), key)



from burlap.files import file # pylint: disable=redefined-builtin
from burlap.tests.functional_tests.base import TestCase
from burlap.common import Satchel

is_file = file.is_file

class _TestSatchel(Satchel):

    name = 'test'

    def configure(self):
        pass

class CommonTests(TestCase):

    def get_test_satchel(self):
        test = _TestSatchel()
        test.genv.hosts = ['localhost']
        test.genv.host_string = test.genv.hosts[0]
        return test

    def test_dryrun(self):

        test = self.get_test_satchel()

        test.dryrun = True
        test.run('touch ~/abc.txt')
        assert not is_file('~/abc.txt')

        test.dryrun = 1
        test.run('touch ~/def.txt')
        assert not is_file('~/def.txt')

        test.dryrun = False
        test.run('touch ~/mno.txt')
        assert not is_file('~/mno.txt')
        test.run('rm -f ~/mno.txt')

        test.dryrun = 0
        test.run('touch ~/xyz.txt')
        assert not is_file('~/xyz.txt')
        test.run('rm -f ~/xyz.txt')

    def test_sudo(self):

        test = self.get_test_satchel()

        all_users = test.run('cut -d: -f1 /etc/passwd')
        print('all users:', all_users)

        ret = (test.sudo('whoami') or '').split('\n')[-1]
        print('ret0:', ret)
        self.assertEqual(ret, 'root')

        if 'travis' in all_users:
            target_user = 'travis'
        else:
            target_user = 'www-data'
        ret = (test.sudo('whoami', user=target_user) or '').split('\n')[-1]
        print('ret1:', ret)
        self.assertEqual(ret, target_user)



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class SnortSatchel(ServiceSatchel):

    name = 'snort'

    #def set_defaults(self):
        #super().set_defaults()

    @property
    def packager_system_packages(self):
        return {
            DEBIAN: [
                'snort',
            ],
            UBUNTU: [
                'snort',
            ],
        }

    @task(precursors=['packager'])
    def configure(self):
        pass

snort = SnortSatchel()



import sys
import traceback

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap import common


class ServiceManagementSatchel(Satchel):

    name = 'service'

    @task
    def pre_deploy(self):
        """
        Runs methods services have requested be run before each deployment.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_pre_deployers.get(service)
            if funcs:
                print('Running pre-deployments for service %s...' % (service,))
                for func in funcs:
                    func()

    @task
    def deploy(self):
        """
        Applies routine, typically application-level changes to the service.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_deployers.get(service)
            if funcs:
                print('Deploying service %s...' % (service,))
                for func in funcs:
                    if not self.dryrun:
                        func()

    @task
    def post_deploy(self):
        """
        Runs methods services have requested be run before after deployment.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            self.vprint('post_deploy:', service)
            funcs = common.service_post_deployers.get(service)
            if funcs:
                self.vprint('Running post-deployments for service %s...' % (service,))
                for func in funcs:
                    try:
                        func()
                    except Exception as e:
                        print('Post deployment error: %s' % e, file=sys.stderr)
                        print(traceback.format_exc(), file=sys.stderr)

    @task
    def pre_db_dump(self):
        """
        Runs methods services that have requested to be run before each
        database dump.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_pre_db_dumpers.get(service)
            if funcs:
                print('Running pre-database dump for service %s...' % (service,))
                for func in funcs:
                    func()

    @task
    def post_db_dump(self):
        """
        Runs methods services that have requested to be run before each
        database dump.
        """
        for service in self.genv.services:
            service = service.strip().upper()
            funcs = common.service_post_db_dumpers.get(service)
            if funcs:
                print('Running post-database dump for service %s...' % (service,))
                for func in funcs:
                    func()


    @task
    def refresh(self):
        self.configure()
        self.deploy()
        self.post_deploy()

    @task
    def configure(self):
        """
        Applies one-time settings changes to the host, usually to initialize the service.
        """
        print('env.services:', self.genv.services)
        for service in list(self.genv.services):
            service = service.strip().upper()
            funcs = common.service_configurators.get(service, [])
            if funcs:
                print('!'*80)
                print('Configuring service %s...' % (service,))
                for func in funcs:
                    print('Function:', func)
                    if not self.dryrun:
                        func()

service = ServiceManagementSatchel()



import os

from fabric.api import runs_once

from burlap.constants import *
from burlap.db import DatabaseSatchel
from burlap.decorators import task, runs_once

class MongoDBSatchel(DatabaseSatchel):
    """
    Represents a MongoDB server.

    Installation follows the instructions outlined in the tutorial:

        https://docs.mongodb.com/manual/tutorial/install-mongodb-on-ubuntu/
    """

    name = 'mongodb'

    def set_defaults(self):
        super().set_defaults()

        self.env.dump_command = 'mongodump -h {db_host}:{db_port} -v --username={db_user} --password={db_password} --gzip --archive={dump_fn}'
        self.env.dump_fn_template = '{dump_dest_dir}/db_mongodb_{SITE}_{ROLE}_{db_name}_$(date +%Y%m%d).archive'

        #self.env.load_command = 'gunzip < {remote_dump_fn} | pg_restore --jobs=8 -U {db_root_username} --format=c --create --dbname={db_name}'
        #self.env.load_command = 'gunzip < {remote_dump_fn} | pg_restore -U {db_root_username} --format=c --create --dbname={db_name}'
        self.env.load_command = 'mongorestore --drop --gzip --noIndexRestore --archive={remote_dump_fn}'
        #https://docs.mongodb.com/v3.0/tutorial/build-indexes-in-the-background/

        self.env.db_port = 9001

        self.env.watchdog_enabled = False
        self.env.watchdog_cron_schedule = None

        self.define_cron_job(
            name='mongomonitor',
            template='etc_crond_mongomonitor',
            # script_path='/etc/cron.d/mongomonitor',
            command='/usr/local/bin/mongomonitor',
            schedule=self.env.watchdog_cron_schedule
        )

    @task
    def install_watchdog(self):
        if not self.env.watchdog_enabled:
            return
        assert self.env.watchdog_cron_schedule, 'No schedule defined for the watchdog!'
        self.install_script(
            local_path='mongodb/mongomonitor',
            remote_path='/usr/local/bin/mongomonitor',
        )
        self.install_cron_job(name='mongomonitor')

    @task
    def configure(self, *args, **kwargs):
        super().configure(*args, **kwargs)
        self.install_watchdog()

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['mongodb-org'],
        }

    @property
    def packager_repositories(self):
        ver = self.os_version
        if ver.type == LINUX:
            if ver.distro == UBUNTU:
                if ver.release == '14.04':
                    d = {
                        APT_SOURCE: [
                            (
                                'deb [ arch=amd64 ] http://repo.mongodb.org/apt/ubuntu trusty/mongodb-org/3.4 multiverse',
                                '/etc/apt/sources.list.d/mongodb-org-3.4.list',
                            ),
                        ],
                        APT_KEY: [
                            ('hkp://keyserver.ubuntu.com:80', '0C49F3730359A14518585931BC711F9BA15703C6'),
                        ],
                    }
                    return d
                if ver.release == '16.04':
                    # https://www.digitalocean.com/community/tutorials/how-to-install-and-secure-mongodb-on-ubuntu-16-04
                    # https://docs.mongodb.com/manual/tutorial/install-mongodb-on-ubuntu/
                    d = {
                        APT_SOURCE: [
                            (
                                'deb [ arch=amd64,arm64 ] http://repo.mongodb.org/apt/ubuntu xenial/mongodb-org/3.4 multiverse',
                                '/etc/apt/sources.list.d/mongodb-org-3.4.list',
                            ),
                        ],
                        APT_KEY: [
                            ('hkp://keyserver.ubuntu.com:80', '0C49F3730359A14518585931BC711F9BA15703C6'),
                        ],
                    }
                    return d
        raise NotImplementedError

    @task
    @runs_once
    def load(self, dump_fn='', prep_only=0, force_upload=0, from_local=0, name=None, site=None, dest_dir=None):
        """
        Restores a database snapshot onto the target database server.

        If prep_only=1, commands for preparing the load will be generated,
        but not the command to finally load the snapshot.
        """

        r = self.database_renderer(name=name, site=site)

        # Render the snapshot filename.
        r.env.dump_fn = self.get_default_db_fn(fn_template=dump_fn, dest_dir=dest_dir)

        from_local = int(from_local)

        prep_only = int(prep_only)

        missing_local_dump_error = r.format('Database dump file {dump_fn} does not exist.')

        # Copy snapshot file to target.
        if self.is_local:
            r.env.remote_dump_fn = dump_fn
        else:
            r.env.remote_dump_fn = '/tmp/' + os.path.split(r.env.dump_fn)[-1]

        if not prep_only and not self.is_local:
            if not self.dryrun:
                assert os.path.isfile(r.env.dump_fn), missing_local_dump_error
            r.pc('Uploading MongoDB database snapshot...')
#                 r.put(
#                     local_path=r.env.dump_fn,
#                     remote_path=r.env.remote_dump_fn)
            r.local('rsync -rvz --progress --no-p --no-g '
                '--rsh "ssh -o StrictHostKeyChecking=no -i {key_filename}" '
                '{dump_fn} {user}@{host_string}:{remote_dump_fn}')

        if self.is_local and not prep_only and not self.dryrun:
            assert os.path.isfile(r.env.dump_fn), missing_local_dump_error

        r.run_or_local(r.env.load_command)

    @task
    def shell(self, name='default', user=None, password=None, root=0, verbose=1, write_password=1, no_db=0, no_pw=0):
        """
        Opens a SQL shell to the given database, assuming the configured database
        and user supports this feature.
        """
        raise NotImplementedError

    @task
    def create(self, **kwargs):
        """
        Creates the target database.
        """
        raise NotImplementedError

    @task
    def drop_views(self, name=None, site=None):
        """
        Drops all views.
        """
        raise NotImplementedError

    @task
    def drop_database(self, name):
        raise NotImplementedError

    @task
    def exists(self, name='default', site=None):
        """
        Returns true if a database with the given name exists. False otherwise.
        """
        raise NotImplementedError

mongodb = MongoDBSatchel()



import os
import socket
from unittest import skipIf

from burlap.tests.functional_tests.base import TestCase

def is_valid_ipv4_address(address):
    try:
        _ = socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            _ = socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:    # not a valid address
        return False
    return True


def is_valid_ipv6_address(address):
    try:
        _ = socket.inet_pton(socket.AF_INET6, address)
    except socket.error:    # not a valid address
        return False
    return True


def is_valid_address(address):
    return is_valid_ipv4_address(address) or is_valid_ipv6_address()


class NetworkTests(TestCase):

    @skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true", "Skipping this test on Travis CI.") #TODO:fix on Travis?
    def test_interfaces_ipv4_addresses(self):

        from burlap.network import interfaces # pylint: disable=import-outside-toplevel
        from burlap.network import address # pylint: disable=import-outside-toplevel

        for interface in interfaces():
            ipv4_address = address(interface)
            assert is_valid_ipv4_address(ipv4_address)

    def test_loopback_interface_exists(self):
        from burlap.network import interfaces # pylint: disable=import-outside-toplevel
        assert 'lo' in interfaces()

    def test_loopback_interface_address(self):
        from burlap.network import address # pylint: disable=import-outside-toplevel
        assert address('lo') == '127.0.0.1'

    @skipIf("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true", "Skipping this test on Travis CI.") #TODO:fix on Travis?
    def test_name_servers(self):
        from burlap.network import nameservers # pylint: disable=import-outside-toplevel
        for address in nameservers():
            assert is_valid_address(address)



import os
import shutil
from subprocess import getoutput

from burlap.constants import *
from burlap.common import get_satchel
from burlap.tests.base import TestCase
from burlap.context import set_cwd


class PipTests(TestCase):

    def test_pip_install(self):
        pip_satchel = get_satchel('pip')

        # Initialize tmp directory.
        d = '/tmp/test_pip_install'
        if os.path.isdir(d):
            shutil.rmtree(d)
        os.makedirs(d)

        try:
            # Install pip requirements.
            with set_cwd(d):

                # Create requirements file.
                os.makedirs('roles/all')
                with open('roles/all/pip-requirements.txt', 'w') as fout:
                    print('PyYAML\n', file=fout)

                # Install without the quiet flag.
                pip_satchel.verbose = 0
                pip_satchel.env.virtualenv_dir = '%s/.env' % d
                pip_satchel.configure()

                self.assertTrue(os.path.isdir(pip_satchel.env.virtualenv_dir))
                ret = getoutput('%s/bin/pip freeze | grep -i yaml' % pip_satchel.env.virtualenv_dir)
                print('pip freeze:\n', ret)
                self.assertIn('PyYAML', ret)

                # Delete the virtualenv.
                shutil.rmtree(os.path.join(d, '.env'))

                # Install with the quiet flag.
                pip_satchel.verbose = GLOBAL_VERBOSE
                pip_satchel.configure()

                self.assertTrue(os.path.isdir('.env'))
                ret = getoutput('.env/bin/pip freeze | grep -i yaml')
                print('pip freeze:\n', ret)
                self.assertIn('PyYAML', ret)

        finally:
            shutil.rmtree(d)



import re
import getpass

from fabric.api import (
    settings,
    runs_once,
)

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task, task_or_dryrun
from burlap.common import str_to_callable

def iter_hostnames():
    from burlap.common import get_hosts_retriever, get_verbose # pylint: disable=import-outside-toplevel

    verbose = get_verbose()

    retriever = get_hosts_retriever()

    hosts = list(retriever(extended=1))
    for _hostname, _data in hosts:
        yield _hostname
#         if _data.ip == env.host_string:
#             yield _hostname
#         elif _data.public_dns_name == env.host_string:
#             yield _hostname


@task_or_dryrun
@runs_once
def list_hostnames():
    for hostname in iter_hostnames():
        print(hostname)


class HostSatchel(Satchel):
    """
    Used for initializing the host.

    Should only be called once, manually on a fresh install to initialize
    the primary user login.

    e.g. fab prod host.initrole host.configure
    """

    name = 'host'

    def set_defaults(self):

        self.env.default_hostname = None
        self.env.default_hosts = []
        self.env.default_user = None
        self.env.default_password = None
        self.env.default_key_filename = None

        self.env.do_initrole = True

        self.env.post_initrole_tasks = []

        self.env.original_user = None
        self.env.original_key_filename = None

        self.env.login_check = False

#         self.env.os_type = None # Linux/Windows/etc
#         self.env.os_distro = None # Ubuntu/Fedora/etc
#         self.env.os_release = None # 12.04/14.04/etc

    def hostname_to_ip(self, hostname):
        r = self.local_renderer
        r.env.hostname = hostname
        return r._local("getent hosts {hostname} | awk '{{ print $1 }}'", capture=True) or ''

    @task
    def is_present(self, host=None):
        """
        Returns true if the given host exists on the network.
        Returns false otherwise.
        """
        r = self.local_renderer
        r.env.host = host or self.genv.host_string
        ret = r._local("getent hosts {host} | awk '{{ print $1 }}'", capture=True) or ''
        if self.verbose:
            print('ret:', ret)
        ret = ret.strip()
        if self.verbose:
            print('Host %s %s present.' % (r.env.host, 'IS' if bool(ret) else 'IS NOT'))
        ip = ret
        ret = bool(ret)
        if not ret:
            return False

        r.env.ip = ip
        with settings(warn_only=True):
            ret = r._local('ping -c 1 {ip}', capture=True) or ''
        packet_loss = re.findall(r'([0-9]+)% packet loss', ret)
#         print('packet_loss:',packet_loss)
        ip_accessible = packet_loss and int(packet_loss[0]) < 100
        if self.verbose:
            print('IP %s accessible: %s' % (ip, ip_accessible))
        return bool(ip_accessible)

    @task
    def purge_keys(self):
        """
        Deletes all SSH keys on the localhost associated with the current remote host.
        """
        r = self.local_renderer
        r.env.default_ip = self.hostname_to_ip(self.env.default_hostname)
        r.env.home_dir = '/home/%s' % getpass.getuser()
        r.local('ssh-keygen -f "{home_dir}/.ssh/known_hosts" -R {host_string}')
        if self.env.default_hostname:
            r.local('ssh-keygen -f "{home_dir}/.ssh/known_hosts" -R {default_hostname}')
        if r.env.default_ip:
            r.local('ssh-keygen -f "{home_dir}/.ssh/known_hosts" -R {default_ip}')

    @task
    def find_working_password(self, usernames=None, host_strings=None):
        """
        Returns the first working combination of username and password for the current host.
        """
        r = self.local_renderer

        if host_strings is None:
            host_strings = []

        if not host_strings:
            host_strings.append(self.genv.host_string)

        if usernames is None:
            usernames = []

        if not usernames:
            usernames.append(self.genv.user)

        for host_string in host_strings:

            for username in usernames:

                passwords = []
                passwords.append(self.genv.user_default_passwords[username])
                passwords.append(self.genv.user_passwords[username])
                passwords.append(self.env.default_password)

                for password in passwords:

                    with settings(warn_only=True):
                        r.env.host_string = host_string
                        r.env.password = password
                        r.env.user = username
                        ret = r._local("sshpass -p '{password}' ssh -o StrictHostKeyChecking=no {user}@{host_string} echo hello", capture=True)
                        #print('ret.return_code:', ret.return_code)
            #             print('ret000:[%s]' % ret)
                        #code 1 = good password, but prompts needed
                        #code 5 = bad password
                        #code 6 = good password, but host public key is unknown

                    if ret.return_code in (1, 6) or 'hello' in ret:
                        # Login succeeded, so we haven't yet changed the password, so use the default password.
                        return host_string, username, password

        raise Exception('No working login found.')

    @task
    def needs_initrole(self, stop_on_error=False):
        """
        Returns true if the host does not exist at the expected location and may need
        to have its initial configuration set.
        Returns false if the host exists at the expected location.
        """

        ret = False

        target_host_present = self.is_present()

        if not target_host_present:
            default_host_present = self.is_present(self.env.default_hostname)
            if default_host_present:
                if self.verbose:
                    print('Target host missing and default host present so host init required.')
                ret = True
            else:
                if self.verbose:
                    print('Target host missing but default host also missing, '
                        'so no host init required.')
#                 if stop_on_error:
#                     raise Exception(
#                         'Both target and default hosts missing! '
#                         'Is the machine turned on and plugged into the network?')
        else:
            if self.verbose:
                print('Target host is present so no host init required.')

        return ret

    @task
    def initrole(self, check=True):
        """
        Called to set default password login for systems that do not yet have passwordless
        login setup.
        """

        if self.env.original_user is None:
            self.env.original_user = self.genv.user

        if self.env.original_key_filename is None:
            self.env.original_key_filename = self.genv.key_filename

        host_string = None
        user = None
        password = None
        if self.env.login_check:
            host_string, user, password = self.find_working_password(
                usernames=[self.genv.user, self.env.default_user],
                host_strings=[self.genv.host_string, self.env.default_hostname],
            )
            if self.verbose:
                print('host.initrole.host_string:', host_string)
                print('host.initrole.user:', user)
                print('host.initrole.password:', password)

#         needs = True
#         if check:
#             needs = self.needs_initrole(stop_on_error=True)
        needs = False

        if host_string is not None:
            self.genv.host_string = host_string
        if user is not None:
            self.genv.user = user
        if password is not None:
            self.genv.password = password

        if not needs:
            return

        assert self.env.default_hostname, 'No default hostname set.'
        assert self.env.default_user, 'No default user set.'

        self.genv.host_string = self.env.default_hostname
        if self.env.default_hosts:
            self.genv.hosts = self.env.default_hosts
        else:
            self.genv.hosts = [self.env.default_hostname]

        self.genv.user = self.env.default_user
        self.genv.password = self.env.default_password
        self.genv.key_filename = self.env.default_key_filename

        # If the host has been reformatted, the SSH keys will mismatch, throwing an error, so clear them.
        self.purge_keys()

        # Do a test login with the default password to determine which password we should use.
#         r.env.password = self.env.default_password
#         with settings(warn_only=True):
#             ret = r._local("sshpass -p '{password}' ssh -o StrictHostKeyChecking=no {user}@{host_string} echo hello", capture=True)
#             print('ret.return_code:', ret.return_code)
# #             print('ret000:[%s]' % ret)
#             #code 1 = good password, but prompts needed
#             #code 5 = bad password
#             #code 6 = good password, but host public key is unknown
#         if ret.return_code in (1, 6) or 'hello' in ret:
#             # Login succeeded, so we haven't yet changed the password, so use the default password.
#             self.genv.password = self.env.default_password
#         elif self.genv.user in self.genv.user_passwords:
#             # Otherwise, use the password or key set in the config.
#             self.genv.password = self.genv.user_passwords[self.genv.user]
#         else:
#             # Default password fails and there's no current password, so clear.
#             self.genv.password = None
#         self.genv.password = self.find_working_password()
#         print('host.initrole,using password:', self.genv.password)

        # Execute post-init callbacks.
        for task_name in self.env.post_initrole_tasks:
            if self.verbose:
                print('Calling post initrole task %s' % task_name)
            satchel_name, method_name = task_name.split('.')
            satchel = self.get_satchel(name=satchel_name)
            getattr(satchel, method_name)()

        print('^'*80)
        print('host.initrole.host_string:', self.genv.host_string)
        print('host.initrole.user:', self.genv.user)
        print('host.initrole.password:', self.genv.password)

    @task
    def reboot_test(self):
        r = self.local_renderer
        r.run('echo before reboot')
        r.reboot(wait=300)
        r.run('echo after reboot')

    def deploy_pre_run(self):

        # If the desired hostname is not present but the default hostname is present,
        # then we assume the host is new or has been reset and needs to be reconfigured.
        if self.env.do_initrole:
            self.initrole()

    def record_manifest(self, *args, **kargs):
        self.initrole()
        return super().record_manifest(*args, **kargs)

    def configure(self):
        # Just a stub. All the magic happens in deploy_pre_run().
        pass

UNKNOWN = '?'

class HostnameSatchel(Satchel):

    name = 'hostname'

    def set_defaults(self):

        self.env.hostnames = {} # {ip: hostname}

        self.env.default_hostnames = {} # {target hostname: original hostname}

        self.env.use_retriever = False

        # The bash command to run to get our public IP as we appear on the WAN.
        self.env.get_public_ip_command = 'wget -qO- http://ipecho.net/plain ; echo'

        self.env.set_hostname_command = 'hostnamectl set-hostname {hostname}'

        self.env.auto_reboot = 1

    def record_manifest(self):
        """
        Returns a dictionary representing a serialized state of the service.
        """
        data = {}
        #data['hostnames'] = sorted(list(set(iter_hostnames())))
        data['hostnames'] = self.env.hostnames
        return data

    def hostname_to_ip(self, hostname):
        r = self.local_renderer
        r.env.hostname = hostname
#         print('self.genv.hosts:',self.genv.hosts)
#         print('self.genv.host_string:',self.genv.host_string)
        ret = r._local("getent hosts {hostname} | awk '{{ print $1 }}'", capture=True) or ''
#         print('ret:', ret)
        return ret

    def iter_hostnames(self):
        """
        Yields a list of tuples of the form (ip, hostname).
        """
        from burlap.common import get_hosts_retriever # pylint: disable=import-outside-toplevel
        if self.env.use_retriever:
            self.vprint('using retriever')
            self.vprint('hosts:', self.genv.hosts)
            retriever = get_hosts_retriever()
            hosts = list(retriever(extended=1))
            for _hostname, _data in hosts:

                # Skip hosts that aren't selected for this run.
                if self.genv.hosts \
                and _data.ip not in self.genv.hosts \
                and _data.public_dns_name not in self.genv.hosts \
                and _hostname not in self.genv.hosts:
                    continue

                assert _data.ip, 'Missing IP.'
                yield _data.ip, _hostname#_data.public_dns_name
        else:
            self.vprint('using default')
            for ip, hostname in self.env.hostnames.items():
                self.vprint('ip lookup:', ip, hostname)
                if ip == UNKNOWN:
                    ip = self.hostname_to_ip(hostname)
                    if not ip and hostname in self.env.default_hostnames:
                        ip = self.hostname_to_ip(self.env.default_hostnames[hostname])
                elif not ip[0].isdigit():
                    ip = self.hostname_to_ip(ip)
                assert ip, 'Invalid IP.'
                yield ip, hostname

    @task
    def get_public_ip(self):
        """
        Gets the public IP for a host.
        """
        r = self.local_renderer
        ret = r.run(r.env.get_public_ip_command) or ''
        ret = ret.strip()
        print('ip:', ret)
        return ret

    @task
    @runs_once
    def configure(self, reboot=1):
        """
        Assigns a name to the server accessible from user space.

        Note, we add the name to /etc/hosts since not all programs use
        /etc/hostname to reliably identify the server hostname.
        """
        r = self.local_renderer
        for ip, hostname in self.iter_hostnames():
            self.vprint('ip/hostname:', ip, hostname)
            r.genv.host_string = ip
            r.env.hostname = hostname
            with settings(warn_only=True):
                r.sudo('echo "{hostname}" > /etc/hostname')
                r.sudo('echo "127.0.0.1 {hostname}" | cat - /etc/hosts > /tmp/out && mv /tmp/out /etc/hosts')
                r.sudo(r.env.set_hostname_command)
                if r.env.auto_reboot and int(reboot):
                    r.reboot()

class HostsFileSatchel(Satchel):
    """
    Manages customizations to /etc/hosts.
    """

    name = 'hostsfile'

    def set_defaults(self):
        self.env.ipdomains = [] # [[ip, domain]]
        self.env.ipdomain_retriever = None

    @task(precursors=['packager'])
    def configure(self, force=0):
        force = int(force)
        last_hosts = set(tuple(_) for _ in (self.last_manifest.ipdomains or []))
        ipdomains = list(self.env.ipdomains or [])

        current_hosts = set(tuple(_) for _ in ipdomains)
        if self.env.ipdomain_retriever:
            self.vprint('Looking up hosts with:', self.env.ipdomain_retriever)
            current_hosts.update(str_to_callable(self.env.ipdomain_retriever)(role=self.genv.ROLE) or [])
        self.vprint('current hosts:', sorted(current_hosts))

        if force:
            added_hosts = current_hosts
            removed_hosts = last_hosts.difference(current_hosts)
        else:
            added_hosts = current_hosts.difference(last_hosts)
            removed_hosts = last_hosts.difference(current_hosts)

        r = self.local_renderer
        if self.verbose:
            print('ipdomains:', sorted(ipdomains))
            print('added_hosts:', sorted(added_hosts))
            print('removed_hosts:', sorted(removed_hosts))

        for _ip, _domain in sorted(added_hosts):
            r.env.ip = _ip
            r.env.domain = _domain
            r.append(filename='/etc/hosts', text='%s %s' % (_ip, _domain), use_sudo=True)

        for _ip, _domain in sorted(removed_hosts):
            r.env.ip = _ip
            r.env.domain = _domain
            r.sed(filename='/etc/hosts', before='%s %s' % (_ip, _domain), after='', use_sudo=True)

class TimezoneSatchel(Satchel):
    """
    Manages setting the system-wide timezone.

    For a list of standard timezones see:

        https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """

    name = 'timezone'

    def set_defaults(self):
        self.env.timezone = 'UTC'

    @task
    def get_current_timezone(self):
        ret = self.sudo('dpkg-reconfigure -f noninteractive tzdata') or ''
        matches = re.findall(r'Local time is now:.*?[0-9]+\s+([A-Z]+)\s+[0-9]+', ret)
        self.vprint('matches:', matches)
        if matches:
            return matches[0]
        return (self.run('date +%Z') or self.env.timezone).strip()

    @task(precursors=['packager'])
    def configure(self):
        r = self.local_renderer
        os_ver = self.os_version
        if os_ver.distro == UBUNTU and os_ver.release >= '16.04':
            r.sudo('timedatectl set-timezone {timezone}')
        else:
            # Old way in Ubuntu <= 14.04.
            r.sudo("sudo sh -c 'echo \"{timezone}\" > /etc/timezone'")
            r.sudo('dpkg-reconfigure -f noninteractive tzdata')

host = HostSatchel()
hostname = HostnameSatchel()
hostsfile = HostsFileSatchel()
timezone = TimezoneSatchel()



"""
Git
===

This module provides low-level tools for managing `Git`_ repositories.  You
should normally not use them directly but rather use the high-level wrapper
:func:`fabtools.require.git.working_copy` instead.

.. _Git: http://git-scm.com/

"""
import os

from fabric.api import run, sudo, hide
from fabric.context_managers import cd

from burlap import Satchel
from burlap.constants import *
from burlap.decorators import task
from burlap.exceptions import AbortDeployment
from burlap.utils import run_as_root
from burlap.files import file # pylint: disable=redefined-builtin

CURRENT_COMMIT = 'current_commit'

class GitSatchel(Satchel):

    name = 'git'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['git'],
        }

    def set_defaults(self):
        self.env.hooks = {
            #path/to/git/repo': ['local/path/to/hook']
        }

    @task
    def install_hooks(self):
        r = self.local_renderer
        for repo_path, hook_paths in r.env.hooks.items():
            if repo_path == '.' and self.is_local:
                repo_path = os.getcwd()
            r.env.repo_path_dot_git = '%s/.git' % repo_path
            assert file.is_dir(r.env.repo_path_dot_git), 'Repo %s does not exist.' % r.env.repo_path_dot_git
            for hook_path in hook_paths:
                r.env.local_hook_path = fqfn = self.find_template(hook_path)
                r.env.remote_hook_path = '%s/.git/hooks/%s' % (repo_path, os.path.split(hook_path)[-1])
                r.put(local_path=r.env.local_hook_path, remote_path=r.env.remote_hook_path)
                r.run_or_local('chmod +x {remote_hook_path}')

    @task
    def uninstall_hooks(self, hooks):
        raise NotImplementedError

    def clone(self, remote_url, path=None, use_sudo=False, user=None):
        """
        Clone a remote Git repository into a new directory.

        :param remote_url: URL of the remote repository to clone.
        :type remote_url: str

        :param path: Path of the working copy directory.  Must not exist yet.
        :type path: str

        :param use_sudo: If ``True`` execute ``git`` with
                         :func:`fabric.operations.sudo`, else with
                         :func:`fabric.operations.run`.
        :type use_sudo: bool

        :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                     with the given user.  If ``use_sudo is False`` this parameter
                     has no effect.
        :type user: str
        """

        cmd = 'git clone --quiet %s' % remote_url
        if path is not None:
            cmd = cmd + ' %s' % path

        if use_sudo and user is None:
            run_as_root(cmd)
        elif use_sudo:
            sudo(cmd, user=user)
        else:
            run(cmd)


    def add_remote(self, path, name, remote_url, use_sudo=False, user=None, fetch=True):
        """
        Add a remote Git repository into a directory.

        :param path: Path of the working copy directory.  This directory must exist
                     and be a Git working copy with a default remote to fetch from.
        :type path: str

        :param use_sudo: If ``True`` execute ``git`` with
                         :func:`fabric.operations.sudo`, else with
                         :func:`fabric.operations.run`.
        :type use_sudo: bool

        :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                     with the given user.  If ``use_sudo is False`` this parameter
                     has no effect.
        :type user: str

        :param name: name for the remote repository
        :type name: str

        :param remote_url: URL of the remote repository
        :type remote_url: str

        :param fetch: If ``True`` execute ``git remote add -f``
        :type fetch: bool
        """
        if path is None:
            raise ValueError("Path to the working copy is needed to add a remote")

        if fetch:
            cmd = 'git remote add -f %s %s' % (name, remote_url)
        else:
            cmd = 'git remote add %s %s' % (name, remote_url)

        with cd(path):
            if use_sudo and user is None:
                run_as_root(cmd)
            elif use_sudo:
                sudo(cmd, user=user)
            else:
                run(cmd)


    def fetch(self, path, use_sudo=False, user=None, remote=None):
        """
        Fetch changes from the default remote repository.

        This will fetch new changesets, but will not update the contents of
        the working tree unless yo do a merge or rebase.

        :param path: Path of the working copy directory.  This directory must exist
                     and be a Git working copy with a default remote to fetch from.
        :type path: str

        :param use_sudo: If ``True`` execute ``git`` with
                         :func:`fabric.operations.sudo`, else with
                         :func:`fabric.operations.run`.
        :type use_sudo: bool

        :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                     with the given user.  If ``use_sudo is False`` this parameter
                     has no effect.
        :type user: str

        :type remote: Fetch this remote or default remote if is None
        :type remote: str
        """

        if path is None:
            raise ValueError("Path to the working copy is needed to fetch from a remote repository.")

        if remote is not None:
            cmd = 'git fetch %s' % remote
        else:
            cmd = 'git fetch'

        with cd(path):
            if use_sudo and user is None:
                run_as_root(cmd)
            elif use_sudo:
                sudo(cmd, user=user)
            else:
                run(cmd)


    def pull(self, path, use_sudo=False, user=None, force=False):
        """
        Fetch changes from the default remote repository and merge them.

        :param path: Path of the working copy directory.  This directory must exist
                     and be a Git working copy with a default remote to pull from.
        :type path: str

        :param use_sudo: If ``True`` execute ``git`` with
                         :func:`fabric.operations.sudo`, else with
                         :func:`fabric.operations.run`.
        :type use_sudo: bool

        :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                     with the given user.  If ``use_sudo is False`` this parameter
                     has no effect.
        :type user: str
        :param force: If ``True``, append the ``--force`` option to the command.
        :type force: bool
        """

        if path is None:
            raise ValueError("Path to the working copy is needed to pull from a remote repository.")

        options = []
        if force:
            options.append('--force')
        options = ' '.join(options)

        cmd = 'git pull %s' % options

        with cd(path):
            if use_sudo and user is None:
                run_as_root(cmd)
            elif use_sudo:
                sudo(cmd, user=user)
            else:
                run(cmd)


    def checkout(self, path, branch="master", use_sudo=False, user=None, force=False):
        """
        Checkout a branch to the working directory.

        :param path: Path of the working copy directory.  This directory must exist
                     and be a Git working copy.
        :type path: str

        :param branch: Name of the branch to checkout.
        :type branch: str

        :param use_sudo: If ``True`` execute ``git`` with
                         :func:`fabric.operations.sudo`, else with
                         :func:`fabric.operations.run`.
        :type use_sudo: bool

        :param user: If ``use_sudo is True``, run :func:`fabric.operations.sudo`
                     with the given user.  If ``use_sudo is False`` this parameter
                     has no effect.
        :type user: str
        :param force: If ``True``, append the ``--force`` option to the command.
        :type force: bool
        """

        if path is None:
            raise ValueError("Path to the working copy is needed to checkout a branch")

        options = []
        if force:
            options.append('--force')
        options = ' '.join(options)

        cmd = 'git checkout %s %s' % (branch, options)

        with cd(path):
            if use_sudo and user is None:
                run_as_root(cmd)
            elif use_sudo:
                sudo(cmd, user=user)
            else:
                run(cmd)

    def get_changed_hooks(self):
        lm = self.last_manifest
        cm = self.current_manifest
        current_hooks = cm['hooks']
        last_hooks = lm.hooks or {}

        hooks_added = {} # {repo: [hooks]}, in current but not last
        for repo_path in current_hooks:
            hooks_added.setdefault(repo_path, [])
            added = set(current_hooks.get(repo_path, [])).difference(last_hooks.get(repo_path, []))
            hooks_added[repo_path].extend(added)

        hooks_removed = {} # {repo: [hooks]}, in last but not current
        for repo_path in last_hooks:
            hooks_removed.setdefault(repo_path, [])
            removed = set(last_hooks.get(repo_path, [])).difference(current_hooks.get(repo_path, []))
            hooks_removed[repo_path].extend(removed)

        return hooks_added, hooks_removed

    @task
    def update_hooks(self):
        added_hooks, removed_hooks = self.get_changed_hooks()
        if removed_hooks:
            self.uninstall_hooks(removed_hooks)
        if added_hooks:
            self.install_hooks()

    @task(precursors=['packager'])
    def configure(self):
        self.update_hooks()


class GitCheckerSatchel(Satchel):
    """
    Ensures the appropriate Git branch is being deployed.
    """

    name = 'gitchecker'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['git'],
            (UBUNTU, '12.04'): ['git'],
            (UBUNTU, '14.04'): ['git'],
            (UBUNTU, '16.04'): ['git'],
        }

    def set_defaults(self):
        self.env.branch = 'master'

    @task
    def check(self):
        r = self.local_renderer
        r.env.branch = r.format(self.env.branch)
        self.vprint('Checking GIT branch...')
        with hide('running', 'stdout', 'stderr', 'warnings'):
            branch_name = self._local('git rev-parse --abbrev-ref HEAD', capture=True).strip()
            if r.env.branch != branch_name:
                raise AbortDeployment(
                    'You\'re trying to deploy branch "%s" but the target role "%s" only accepts branch "%s".' \
                        % (branch_name, self.genv.ROLE, r.env.branch))

    def record_manifest(self):
        self.check()
        return super().record_manifest()

    @task
    def configure(self):
        pass


class GitTrackerSatchel(Satchel):
    """
    Tracks changes between Git commits.
    """

    name = 'gittracker'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: ['git'],
        }

    def set_defaults(self):
        self.env.callbacks = []

    def get_logs_between_commits(self, old_commit, new_commit):
        """
        Retrieve commit messages for all commits between the given commit hashes.

        This uses an asymmetric diff, so will not return logs for commits descending from old_commit but not new_commit.

        Returns:
            str: git log output, containing full commit hashes and messages, delimited by newlines
        """
        ret = self.local('git --no-pager log --pretty=oneline %s..%s' % (old_commit, new_commit), capture=True)
        if self.verbose:
            print(ret)
        return str(ret)

    @task
    def get_current_commit(self):
        """
        Retrieves the git commit number of the current head branch.
        """
        with hide('running', 'stdout', 'stderr', 'warnings'):
            s = str(self.local('git rev-parse HEAD', capture=True))
            self.vprint('current commit:', s)
            return s

    def record_manifest(self):
        """
        Called after a deployment to record any data necessary to detect changes
        for a future deployment.
        """
        manifest = super().record_manifest()
        manifest[CURRENT_COMMIT] = self.get_current_commit()
        return manifest

    @task(precursors=['packager', 'pip', 'tarball', 'dj'])
    def configure(self):
        for callback in self.env.callbacks:
            satchel_name, method_name = callback.split('.')
            satchel = self.get_satchel(satchel_name)
            method = getattr(satchel, method_name)
            method()

git = GitSatchel()
gitchecker = GitCheckerSatchel()
gittracker = GitTrackerSatchel()



import os
import re

import burlap
from burlap import ContainerSatchel
from burlap.constants import *
from burlap.decorators import task

fabfile_template = os.path.join(
    os.path.dirname(burlap.__file__),
    'templates',
    'burlap',
    'fabfile.py.template',
)

def md(d):
    if os.path.isdir(d):
        return
    os.makedirs(d)

def to_camelcase(value):
    value = re.sub(r'[^a-zA-Z0-9]+', ' ', value).strip()
    return ''.join(x.capitalize() for x in value.split(' '))

def init_dj(project_name, default_roles, virtualenv_dir='.env', version=None, **kwargs):

    site_name = project_name

    print('Installing Django...')
    if version:
        os.system('%s/bin/pip install Django==%s' % (virtualenv_dir, version))
    else:
        os.system('%s/bin/pip install Django' % virtualenv_dir)

    print('Initializing Django project...')
    if not os.path.isdir('src/%s' % site_name):
        print('Initializing base django project...')
        os.system('. %s/bin/activate; django-admin.py startproject %s src; deactivate' % (virtualenv_dir, site_name,))
        _settings_fn = os.path.abspath('src/%s/settings.py' % project_name)
        _content = open(_settings_fn, 'r').read()
        _sites = '''SITE_{name_upper} = "{name_lower}"
SITES = (
    SITE_{name_upper},
)
'''.format(
            name_upper=project_name.upper(),
            name_lower=project_name.lower(),
        )
        _top = []
        for _role in default_roles:
            _top.append("ROLE_%s = '%s'" % (_role.upper(), _role.lower()))
        _top.append('ROLES = (')
        for _role in default_roles:
            _top.append("    ROLE_%s," % (_role.upper(),))
        _top.append(')')
        _index = _content.find('"""\n\n')+4

        bottom_args = dict(
            app_name=project_name,
            app_name_title=project_name.title() + ' Administration',
            app_name_simple=project_name.title()
        )
        _bottom = '''
PROJECT_DIR = os.path.abspath(os.path.join(os.path.split(__file__)[0], '..', '..'))

STATIC_ROOT = os.path.join(PROJECT_DIR, 'static')

MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')
MEDIA_URL = '/media/'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)
TEMPLATE_DIRS = (
    '%s/src/{app_name}/templates' % PROJECT_DIR,
)
# https://docs.djangoproject.com/en/1.11/ref/settings/#templates
TEMPLATES = [
    {{
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': TEMPLATE_DIRS,
        'APP_DIRS': True,
        'OPTIONS': {{
            #'loaders': TEMPLATE_LOADERS, # Unnecessary if we're using APP_DIRS.
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        }},
    }},
]
ADMIN_TITLE = '{app_name_title}'
ADMIN_TITLE_SIMPLE = '{app_name_simple}'
'''.format(**bottom_args)
        open(_settings_fn, 'w').write(_content[:_index]+_sites+('\n'.join(_top))+_content[_index:]+_bottom)

    print('Creating Django helper scripts...')
    open('src/manage', 'w').write('''#!/bin/bash
# Helper script for ensuring we use the Python binary in our local
# virtual environment when calling management commands.
# Otherwise, we'd have to always run `. ../.env/bin/activate`, which can be
# annoying.
# Be sue to run `fab <role> pip.init` first in order to setup
# the target role's Python virtual environment.
DIR=`dirname $0`;
cd $DIR;
../.env/bin/python manage.py $@''')
    open('src/runserver', 'w').write('''#!/bin/bash
# Helper script for running the local dev server, ensuring
# our virtual environment is used.
#set -e
#script_dir=`dirname $0`
#cd $script_dir
if [ -z "$PORT" ]; then
export PORT=8111
fi
if [ -z "$ROLE" ]; then
export ROLE=dev
fi
. ~/.bash_aliases
./manage runserver localhost:$PORT''')
    open('src/shell', 'w').write(r'''#!/bin/bash
# Creates a local PIP-aware shell.
#set -e
if [ $_ == $0 ]
then
echo "Please source this script. Do not execute."
exit 1
fi
#script_dir=`dirname $0`
#cd $script_dir
. .env/bin/activate
PS1="\u@\h:\W(fab)\$ "''')

    md('media')
    md('static')

    os.system('chmod +x src/shell')
    os.system('chmod +x src/manage')
    os.system('chmod +x src/runserver')

    # Create the primary app for containing models/urls/views.
    if not os.path.isdir('src/%s' % project_name):
        os.system('cd src; ./manage startapp %s' % (project_name,))

    os.system('cd src; ./manage syncdb')

class ProjectSatchel(ContainerSatchel):

    name = 'project'

    def set_defaults(self):
        pass

    def update_settings(self, d, role, path='roles/{role}/settings.yaml'):
        """
        Writes a key/value pair to a settings file.
        """
        try:
            import ruamel.yaml # pylint: disable=import-outside-toplevel
            load_func = ruamel.yaml.round_trip_load
            dump_func = ruamel.yaml.round_trip_dump
        except ImportError:
            print('Warning: ruamel.yaml not available, reverting to yaml package, possible lost of formatting may occur.')
            import yaml # pylint: disable=import-outside-toplevel
            load_func = yaml.load
            dump_func = yaml.dump
        settings_fn = path.format(role=role)
        data = load_func(open(settings_fn), Loader=yaml.SafeLoader)
        data.update(d)
        settings_str = dump_func(data)
        open(settings_fn, 'w').write(settings_str)

    @task
    def create_skeleton(self, project_name, roles='', components='', pip_requirements='', virtualenv_dir='.env', **kwargs):

        assert project_name, 'Specify project name.'
        site_name = project_name

        app_name = project_name

        default_roles = [_ for _ in roles.split(',') if _.strip()]
        default_components = [_.strip().lower() for _ in components.split(',') if _.strip()]

        print('Creating folders...')
        md('roles/all')
        for _role in default_roles:
            md('roles/%s' % _role)
        md('src')

        print('Creating roles...')
        open('roles/all/settings.yaml', 'w').write(
            self.render_to_string(
                'burlap/all_settings.yaml.template',
                extra=dict(project_name=project_name, site_name=site_name, app_name=app_name)))
        for _role in default_roles:
            open('roles/%s/settings.yaml' % _role, 'w').write(
                self.render_to_string(
                    'burlap/role_settings.yaml.template',
                    extra=dict(project_name=project_name, site_name=site_name, role=_role)))

        default_packages = pip_requirements.split(',') if pip_requirements else []
        if default_packages:
            open('roles/all/pip-requirements.txt', 'w').write('\n'.join(default_packages))

        print('Adding global apt-requirements.txt...')
        open('roles/all/apt-requirements.txt', 'w').write('')

        print('Adding fabfile...')
        content = open(fabfile_template, 'r').read()
        content = content.format(project_name=project_name)
        open('fabfile.py', 'w').write(content.strip()+'\n')

        print('Initializing local development virtual environment...')
        os.system('python3 -m venv %s' % virtualenv_dir)
        if default_packages:
            os.system('. %s/bin/activate' % virtualenv_dir)
            for package in default_packages:
                cmd = 'pip install %s' % package
                print('cmd:', cmd)
                assert not os.system(cmd)
            os.system('deactivate')

        # Install burlap dependencies.
        burlap_pip_requirements = os.path.join(os.path.dirname(burlap.__file__), 'fixtures/requirements.txt')
        print('burlap_pip_requirements:', burlap_pip_requirements)
        assert os.path.exists(burlap_pip_requirements), 'Missing requirements file: %s' % burlap_pip_requirements
        os.system('. %s/bin/activate' % virtualenv_dir)
        for package in open(burlap_pip_requirements, 'r').readlines():
            if not package.strip():
                continue
            cmd = 'pip install %s' % package
            print('cmd:', cmd)
            assert not os.system(cmd)
        os.system('deactivate')

        print('Adding bash setup...')
        open('setup.bash', 'w').write(self.render_to_string('burlap/setup.bash.template'))

        print('Adding gitignore...')
        open('.gitignore', 'w').write(self.render_to_string('burlap/gitignore.template'))

        args = kwargs.copy()
        args['project_name'] = project_name
        args['roles'] = roles
        args['default_roles'] = default_roles
        args['components'] = components
        args['pip_requirements'] = pip_requirements
        args['virtualenv_dir'] = virtualenv_dir
        for component in default_components:
            print('Setting up component %s...' % component)
            # Get component-specific settings.
            component_kwargs = dict(args)
            for _k, _v in kwargs.items():
                _key = component+'_'
                if _k.startswith(_key):
                    component_kwargs[_k[len(_key):]] = _v
                    del component_kwargs[_k]
            print('component_kwargs:', component_kwargs)
            try:
                globals()['init_%s' % component](**component_kwargs)
            except KeyError:
                pass

        print('='*80)
        print()
        print('Skeleton created for project %s!' % (project_name.title(),))
        print()

    @task
    def add_roles(self, roles):
        for role in roles:
            _role = role.strip().lower()
            fn = 'roles/%s/settings.yaml' % _role
            if os.path.isfile(fn):
                continue
            fn_dir = os.path.split(fn)[0]
            if not os.path.isdir(fn_dir):
                os.makedirs(fn_dir)
            open(fn, 'w').write(
                self.render_to_string('burlap/role_settings.yaml.template', extra=dict(role=_role)))
            print('Added role %s!' % role)

    @task
    def create_satchel(self, name):
        name_simple = re.sub(r'[^a-z0-9]+', '', name.lower())
        content = self.render_to_string(
            'burlap/satchel.py.template',
            extra=dict(
                name_camelcase=to_camelcase(name),
                name_simple=name_simple,
            ))
        if not os.path.isdir('satchels'):
            os.makedirs('satchels')
            os.system('touch satchels/__init__.py')
        satchel_fn = 'satchels/%s.py' % name_simple
        open(satchel_fn, 'w').write(content.strip()+'\n')
        print('Wrote %s.' % satchel_fn)

project = ProjectSatchel()



from burlap import ServiceSatchel
from burlap.constants import *
from burlap.decorators import task

class PostfixSatchel(ServiceSatchel):

    name = 'postfix'

    @property
    def packager_system_packages(self):
        return {
            UBUNTU: [
                'postfix',
                'mailutils',
                'libsasl2-2',
                'ca-certificates',
                'libsasl2-modules',
                'nano',
            ],
        }

    def set_defaults(self):
        self.env.service_commands = {
            START:{
                UBUNTU: 'service postfix start',
            },
            STOP:{
                UBUNTU: 'service postfix stop',
            },
            DISABLE:{
                UBUNTU: 'chkconfig postfix off',
            },
            ENABLE:{
                UBUNTU: 'chkconfig postfix on',
            },
            RESTART:{
                UBUNTU: 'service postfix restart; sleep 5',
            },
            STATUS:{
                UBUNTU: 'service postfix status',
            },
        }
        self.env.domain = '?'
        self.env.enabled = False
        self.env.mailrc_fn = '~/.mailrc'

    @task(precursors=['packager', 'user'])
    def configure(self):
        r = self.local_renderer

        if self.env.enabled:

            # Ensure any previous mail setups are wiped clean.
            # Note, this configuration is not compatible with mailx or sendmail.
            r.sudo('if [ -e "{mailrc_fn}" ]; then rm "{mailrc_fn}"; fi')
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq purge postfix mailutils mailx sendmail')

            r.sudo('debconf-set-selections <<< "postfix postfix/mailname string {domain}"')
            r.sudo('debconf-set-selections <<< "postfix postfix/main_mailer_type string \'Internet Site\'"')
            r.sudo('DEBIAN_FRONTEND=noninteractive apt-get -yq install postfix mailutils libsasl2-2 ca-certificates libsasl2-modules nano')

            remote_path = r.put(
                local_path='etc_postfix_main.cf',
                remote_path='/etc/postfix/main.cf', use_sudo=True)

            r.sudo('cat /etc/ssl/certs/Thawte_Premium_Server_CA.pem | tee -a /etc/postfix/cacert.pem')

            # Note, ensure Gmail account is using 2-step verification: https://accounts.google.com/SmsAuthConfig
            # Note, ensure Gmail account is using an app password: https://security.google.com/settings/security/apppasswords?pli=1
            remote_path = r.put(
                local_path='etc_postfix_sasl_sasl_passwd',
                remote_path='/etc/postfix/sasl/sasl_passwd', use_sudo=True)

            r.sudo('chmod 400 /etc/postfix/sasl/sasl_passwd')
            r.sudo('chown -R root:root /etc/postfix/sasl')
            r.sudo('postmap /etc/postfix/sasl/sasl_passwd')
            r.sudo('update-rc.d postfix defaults')
            r.sudo('service postfix start')
            r.sudo('/usr/sbin/postfix reload')

            # Test with:
            #echo "Test mail from postfix" | mail -s "Test Postfix" user@domain

        else:
            self.disable()
            self.stop()

PostfixSatchel()



import os
import sys
import unittest
try:
    from commands import getstatusoutput
except ImportError:
    from subprocess import getstatusoutput

from burlap.common import set_state, get_state, clear_state, init_env, default_env, env, all_satchels, get_dryrun, set_dryrun, get_verbose, set_verbose, \
    is_callable


def clear_runs_once(func):
    if hasattr(func, 'return_value'):
        print('clearing runs_once on %s' % func)
        print('return_value:', func.return_value)
        print('return_value:', func.wrapped.return_value)
        print('__dict__:', func.__dict__)
        # Fabric wraps function using a class that passes through get/hasattr
        # so we have to try deleting the attribute on a few levels.
        try:
            delattr(func, 'return_value')
        except AttributeError:
            pass
        try:
            delattr(func.wrapped, 'return_value')
        except AttributeError:
            pass
        assert not hasattr(func, 'return_value'), 'Unable to clear runs_once on %s' % func


class TestCase(unittest.TestCase):

    test_name_fout = sys.stdout

    test_name_format = '\n{bar}\nRunning test: {name}\n{bar}\n'

    # These keys will not be cleared between tests.
    # This is useful for keeping the Vagrant login, which is only set once before all the tests are run.
    keep_env_keys = []

    def getstatusoutput(self, cmd):
        print(cmd)
        status, output = getstatusoutput(cmd)
        print('output:', output)
        return status, output

    def get_keep_env_keys(self):
        return list(self.keep_env_keys)

    def clear_env(self):
        keep_env_keys = set(self.get_keep_env_keys())
        for k, v in list(env.items()):
            if k in keep_env_keys:
                continue
            del env[k]

    def update_env(self, d):
        keep_env_keys = set(self.get_keep_env_keys())
        for k, v in default_env.items():
            if k in keep_env_keys:
                continue
            env[k] = v

    def setUp(self):
        from burlap import deploy, manifest
        from burlap.deploy import deploy as deploy_satchel

        # Always print the current test name before the test.
        # _, columns = map(int, os.popen('stty size', 'r').read().split()) # TODO:fix? broke in Ubuntu16+Python3
        columns = 80
        kwargs = dict(
            bar='#'*columns,
            name=self._testMethodName,
        )
        print(self.test_name_format.format(**kwargs), file=self.test_name_fout)

        # Save fabric state.
        self._env = env.copy()

        # Reset fabric state.
        # print('setUp: initializing env...')
        init_env()
        #deploy_init_env()

        if not env.host_string:
            env.host_string = 'localhost'
            env.hosts = [env.host_string]

        # Save cwd.
        self._cwd = os.getcwd()
        # print('cwd:', self._cwd)

        # Save burlap state.
        # print('setUp: Saving burlap state...')
        self._burlap_state = get_state()

        self._dryrun = get_dryrun()
        self._verbose = get_verbose()

        # Clear runs_once on legacy runs_once methods.
        # print('setUp: Clearing runs_once methods...')
        modules = [deploy, deploy_satchel, manifest]
        for module in modules:
            # print('setUp: Checking module:', module)
            for name in dir(module):
                # print('setUp: Checking name:', name)
                if not is_callable(module, name):
                    continue
                func = getattr(module, name)
                # print('clearing:', func)
                clear_runs_once(func)

        # Clear runs_once on our custom runs_once methods.
        # print('setUp: Clearing custom runs_once methods...')
        from burlap.common import runs_once_methods
        for meth in runs_once_methods:
            clear_runs_once(func)

        # Ensure all satchels re-push all their local variables back into the global env.
        # print('setUp: Clearing satchels...')
        for satchel in all_satchels.values():
            satchel.register()
            satchel.clear_caches()

        # Set satchel variables that should be customized just for unittests.
        # For example, so we can run unittests locally, we want to change the default burlap paths so they don't conflict with the defaults,
        # in case we're using burlap to deploy locally.
        deploy_satchel.env.lockfile_path = '/tmp/burlap_unittests/deploy.lock'
        deploy_satchel.env.data_dir = '/tmp/burlap_unittests'

        # Since these tests are automated, if we ever get a prompt, we should immediately fail,
        # because no action should ever be user-interactive.
        env.abort_on_prompts = True
        env.always_use_pty = False

        # print('setUp: Purging deployments...')
        deploy_satchel.purge()

        super().setUp()

    def tearDown(self):

        set_dryrun(self._dryrun)
        set_verbose(self._verbose)

        # Restore CWD.
        os.chdir(self._cwd)

        # Restore fabric state.
        self.clear_env()
        self.update_env(self._env)

        # Restore burlap state.
        clear_state()
        set_state(self._burlap_state)

        super().tearDown()
